
/***************************************************************************
 *   This file is part of Aspect, a simple PEC tool.                       *
 *                                                                         *
 *   Copyright (C) 2006-2007 by Wolfgang Hoffmann <woho@woho.de>           *
 *                                                                         *
 *   This program is free software, licensed under the GPL v2.             *
 *   See the file COPYING for more details.                                *
 ***************************************************************************/


#include "mcucmd.h"
#include "mcuserial.h"
#include "log.h"

#include <QtCore>




/********************************************************************/
/*!
    \class IMcuCmd
    \brief Base for MCU serial communication command

    This class is an approach to generically deal with the
    variety of different command and answer types.
*/


/*!
    \fn bool IMcuCmd::flush() const
    Return \c true if pending bytes for read from serial port
    should be flushed before sending command.

    The default implementation returns \c true.
*/


/*!
    \fn QByteArray IMcuCmd::command() const
    Get raw command string for submission to the MCU.
*/


/*!
    \fn int IMcuCmd::receive(const QByteArray &aAnswerFragment)

    McuSerial uses this method to hand over raw characters
    received from the MCU in \a aAnswerFragment.

    The return value indicates how much more answer
    characters are expected. It can change between
    subsequent calls of receive(), since some commands
    return a status character and, if status is not ok,
    additional descriptive characters. A return value of
    \c 0 indicates that no more characters are expected.
*/


/*!
    \fn QString IMcuCmd::name() const
    Return short human-readable name of command for debugging.
*/





/********************************************************************/
/*!
    \class McuAnsHelper
    \brief Convenience class for parsing answer bytearrays
*/

class McuAnsHelper
    {
public:
    McuAnsHelper(const QByteArray &aAns)
        : m_aAns(aAns) {};
    int byte(int nIx) const
        { return (int)(unsigned char)m_aAns[nIx]; };
    int word(int nIx) const
        { return (int)(unsigned char)m_aAns[nIx] +
            (((int)(unsigned char)m_aAns[nIx+1]) << 8); };
private:
    QByteArray m_aAns;
    };


/********************************************************************/
/*!
    \class McuCmdQueryUnlimited
    \brief Generic base for querying undelimited strings
*/

class McuCmdQueryUnlimited : public IMcuCmd
    {
public:
    McuCmdQueryUnlimited(const QByteArray &aCmd, const QString &qsName)
        : m_aCmd(aCmd), m_aAns(), m_qsName(qsName) {};
    QByteArray answer() const { return m_aAns; };

private:
    QByteArray command() const { return m_aCmd; };
    int receive(const QByteArray &aAnswerFragment);
    QString name() const { return m_qsName; };

protected:
    QByteArray m_aCmd;
    QByteArray m_aAns;
    QString m_qsName;
    };


int McuCmdQueryUnlimited::receive(const QByteArray &aAnswerFragment)
    {
    // wait for at least 1 answer character
    // after that, keep receiving until no further characters are available
    m_aAns += aAnswerFragment;
    if (m_aAns.isEmpty())
        return 1;
    else if (aAnswerFragment.isEmpty())
        return 0;
    else
        return 1000;
    }




/********************************************************************/
/*!
    \class McuCmdQueryFixed
    \brief Generic base for querying fixed lenght data
*/

class McuCmdQueryFixed : public IMcuCmd
    {
public:
    McuCmdQueryFixed(const QByteArray &aCmd, int nAnsLen,
        const QString &qsName, bool bPreFlush = true)
        : m_aCmd(aCmd), m_nAnsLen(nAnsLen), m_aAns(),
          m_qsName(qsName), m_bFlush(bPreFlush) {};
    QByteArray answer() const { return m_aAns; };

private:
    bool flush() const { return m_bFlush; };
    QByteArray command() const { return m_aCmd; };
    int receive(const QByteArray &aAnswerFragment);
    QString name() const { return m_qsName; };

protected:
    QByteArray m_aCmd;
    int m_nAnsLen;
    QByteArray m_aAns;
    QString m_qsName;
    bool m_bFlush;
    };


int McuCmdQueryFixed::receive(const QByteArray &aAnswerFragment)
    {
    m_aAns += aAnswerFragment;
    return m_nAnsLen - m_aAns.length();
    }




/********************************************************************/
/*!
    \class McuDataFirmware
*/

bool McuDataFirmware::mcuGet(McuSerial &rPort)
    {
    m_qsFirmware = QString();
    McuCmdQueryUnlimited oCmd(
        QString::fromUtf8("#:GV#").toLatin1().constData(),
        QString::fromUtf8("GetFirmware"));
    if (!rPort.send(&oCmd))
        return false;
    m_qsFirmware = QString::fromLatin1(oCmd.answer().constData());
    return true;
    }




/********************************************************************/
/*!
    \class McuDataState
*/

bool McuDataState::mcuGet(McuSerial &rPort)
    {
    m_aState.clear();
    McuCmdQueryFixed oCmd1(
        QString::fromUtf8("<#:VX#").toLatin1().constData(), 12,
        QString::fromUtf8("GetState1"));
    McuCmdQueryFixed oCmd2(
        QString::fromUtf8(">#:VX#").toLatin1().constData(), 6,
        QString::fromUtf8("GetState1"));
    if ((!rPort.send(&oCmd1)) || (!rPort.send(&oCmd2)))
        return false;
    m_aState = oCmd1.answer() + oCmd2.answer();
    return true;
    }

QString McuDataState::backlash() const
    {
    McuAnsHelper oAns(m_aState);
    int nRaBacklashMethod = oAns.byte(0);
    int nRaBacklashVal = oAns.word(1);
    int nDecBacklashMethod = oAns.byte(3);
    int nDecBacklashVal = oAns.word(4);
    return QString::fromUtf8("RA %1, %2; Dec %3, %4")
        .arg(QString::fromUtf8(
            (nRaBacklashMethod == 0xf0)? "-1 (finish Left)":
            (nRaBacklashMethod == 0x00)?  "0 (none)":
            (nRaBacklashMethod == 0x0f)?  "1 (finish Right)": "?"))
        .arg(nRaBacklashVal)
        .arg(QString::fromUtf8(
            (nDecBacklashMethod == 0xf0)? "-1 (finish Down)":
            (nDecBacklashMethod == 0x00)?  "0 (none)":
            (nDecBacklashMethod == 0x0f)?  "1 (finish Up)": "?"))
        .arg(nDecBacklashVal);
    }

QString McuDataState::guideRate() const
    {
    // guiding rate (and paddle 2x)
    switch (McuAnsHelper(m_aState).byte(6))
        {
        case 0x01: return QString::fromUtf8("1.33x");
        case 0x02: return QString::fromUtf8("1.66x");
        case 0x03: return QString::fromUtf8("2.0x");
        default: return QString::fromUtf8("???");
        }
    }

QString McuDataState::paddle8Rate() const
    {
    // center rate (paddle 8x)
    switch (McuAnsHelper(m_aState).byte(7))
        {
        case 0x07: return QString::fromUtf8("4x");
        case 0x08: return QString::fromUtf8("8x");
        case 0x09: return QString::fromUtf8("16x");
        case 0x0a: return QString::fromUtf8("20x");
        case 0x0b: return QString::fromUtf8("27x");
        case 0x0c: return QString::fromUtf8("40x");
        default: return QString::fromUtf8("???");
        }
    }

QString McuDataState::paddle16Rate() const
    {
    // slew rate (paddle 16x)
    switch (McuAnsHelper(m_aState).byte(8))
        {
        case 0x07: return QString::fromUtf8("4x");
        case 0x08: return QString::fromUtf8("8x");
        case 0x09: return QString::fromUtf8("16x");
        case 0x0a: return QString::fromUtf8("20x");
        case 0x0b: return QString::fromUtf8("27x");
        case 0x0c: return QString::fromUtf8("40x");
        default: return QString::fromUtf8("???");
        }
    }

QString McuDataState::gotoRate() const
    {
    // goto rate
    switch (McuAnsHelper(m_aState).byte(9))
        {
        case 0x07: return QString::fromUtf8("4x");
        case 0x08: return QString::fromUtf8("8x");
        case 0x09: return QString::fromUtf8("16x");
        case 0x0a: return QString::fromUtf8("20x");
        case 0x0b: return QString::fromUtf8("27x");
        case 0x0c: return QString::fromUtf8("40x");
        default: return QString::fromUtf8("???");
        }
    }

QString McuDataState::orientation() const
    {
    McuAnsHelper oAns(m_aState);
    return QString::fromUtf8("RA %1, Orientation %2")
        .arg(QString::fromUtf8(
            ((oAns.byte(10) & 0x0f) == 0x00)? "Original": "Reverse"))
        .arg(QString::fromUtf8(
            ((oAns.byte(10) & 0xf0) == 0x00)? "Left": "Right"));
    }

QString McuDataState::trackingRate() const
    {
    switch (McuAnsHelper(m_aState).byte(11))
        {
        case 0: return QString::fromUtf8("siderial");
        case 1: return QString::fromUtf8("solar");
        case 2: return QString::fromUtf8("lunar");
        case -1: return QString::fromUtf8("off");
        default: return QString::fromUtf8("???");
        }
    }

int McuDataState::gearboxReduction() const
    {
    return McuAnsHelper(m_aState).word(12);
    }

int McuDataState::wormRatio() const
    {
    return McuAnsHelper(m_aState).word(14);
    }

int McuDataState::wormTeeth() const
    {
    return McuAnsHelper(m_aState).word(16);
    }




/********************************************************************/
/*!
    \class McuDataPecState
*/

bool McuDataPecState::mcuGet(McuSerial &rPort)
    {
    m_aState.clear();
    McuCmdQueryFixed oCmd(
        QString::fromUtf8("#:VS#").toLatin1().constData(), 10,
        QString::fromUtf8("GetPecState"));
    m_oTimeBeg = QTime::currentTime();
    if (!rPort.send(&oCmd))
        return false;
    //m_oTimeEnd = QTime::currentTime();
    m_aState = oCmd.answer();
    return true;
    }

QString McuDataPecState::state() const
    {
    McuAnsHelper oAns(m_aState);
    int nState = oAns.byte(5);
    bool bOk = ((nState & 0x80) != 0);
    bool bOnHandbox = ((nState & 0x01) != 0);
    bool bOnSerial = ((nState & 0x02) != 0);
    bool bTraining = ((nState & 0x04) != 0);
    return QString::fromUtf8("%1, %2%3%4 %5 %6 %7 %8 %9 %10 %11 %12")
        .arg(QString::fromUtf8(bOk? "ok": "not ok"))
        .arg(QString::fromUtf8(bTraining? "teach": (bOnHandbox || bOnSerial)? "on": "off"))
        .arg(QString::fromUtf8(bOnHandbox? " hand": ""))
        .arg(QString::fromUtf8(bOnSerial? " serial": ""))
        .arg(oAns.byte(0), 2, 16, QLatin1Char('0'))
        .arg(oAns.word(1), 4, 16, QLatin1Char('0'))
        .arg(oAns.word(3), 4, 16, QLatin1Char('0'))
        .arg(nState, 2, 16, QLatin1Char('0'))
        .arg(oAns.byte(6), 2, 16, QLatin1Char('0'))
        .arg(oAns.byte(7), 2, 16, QLatin1Char('0'))
        .arg(oAns.byte(8), 2, 16, QLatin1Char('0'))
        .arg(oAns.byte(9), 2, 16, QLatin1Char('0'));
    }

double McuDataPecState::wormPos() const
    {
    McuAnsHelper oAns(m_aState);
    int nPecDiv = oAns.word(1);
    int nPecPos = oAns.word(3);
    int nPecPoint = ((oAns.byte(6) & 0x3f) << 2) + (oAns.byte(7) & 0x3);
    return (double)nPecPoint + (double)nPecPos / (double)nPecDiv;
    }




/********************************************************************/
/*!
    \class McuDataPecTable
*/

McuDataPecTable::McuDataPecTable()
    {
    m_nFactor = 1;
    for (int nRow = 0; nRow < c_nRowNum; nRow++)
        m_anValue[nRow] = 0;
    }

bool McuDataPecTable::mcuGetFactor(McuSerial &rPort)
    {
    McuCmdQueryFixed oCmd(
        QString::fromUtf8("#:VS#").toLatin1().constData(), 10,
        QString::fromUtf8("GetPecFactor"));
    if (!rPort.send(&oCmd))
        return false;
    m_nFactor = McuAnsHelper(oCmd.answer()).byte(9);
    return true;
    }

bool McuDataPecTable::mcuGetValue(McuSerial &rPort, int nRow)
    {
    if ((nRow < 0) || (nRow >= c_nRowNum))
        return false;
    McuCmdQueryFixed oCmd(
        QString::fromUtf8("#%1:VR#").arg(nRow).toLatin1().constData(), 1,
        QString::fromUtf8("GetPecValue"));
    if (!rPort.send(&oCmd))
        return false;
    m_anValue[nRow] = McuAnsHelper(oCmd.answer()).byte(0) - '0';
    return true;
    }


bool McuDataPecTable::mcuSetFactor(McuSerial &rPort) const
    {
    McuCmdQueryFixed oCmd1(
        QString::fromUtf8("#%1p#").arg(m_nFactor).toLatin1().constData(), 0,
        QString::fromUtf8("SetPecFactor"));
    McuCmdQueryFixed oCmd2(
        QString::fromUtf8("Y").toLatin1().constData(), 0,
        QString::fromUtf8("SaveToEEPROM"));
    if ((!rPort.send(&oCmd1)) || (!rPort.send(&oCmd2)))
        return false;
    return true;
    }

bool McuDataPecTable::mcuSetValue(McuSerial &rPort, int nRow) const
    {
    if ((nRow < 0) || (nRow >= c_nRowNum))
        return false;
    int nVal = m_anValue[nRow];
    McuCmdQueryFixed oCmd(
        QString::fromUtf8("#%1:XR%2#").arg(nRow).arg(nVal).toLatin1().constData(), 0,
        QString::fromUtf8("SetPecValue"));
    if (!rPort.send(&oCmd))
        return false;
    return true;
    }

bool McuDataPecTable::mcuSave(McuSerial &rPort) const
    {
    McuCmdQueryFixed oCmd(
        QString::fromUtf8("#:XYR#").toLatin1().constData(), 0,
        QString::fromUtf8("SavePecTable"));
    if (!rPort.send(&oCmd))
        return false;
    return true;
    }

bool McuDataPecTable::setFactor(int nFactor)
    {
    if ((nFactor < 0) || (nFactor > 9))
        return false;
    m_nFactor = nFactor;
    return true;
    }


bool McuDataPecTable::setValue(int nRow, int nValue)
    {
    if ((nRow < 0) || (nRow >= c_nRowNum))
        return false;
    if ((nValue < 0) || (nValue > 3))
        return false;
    m_anValue[nRow] = nValue;
    return true;
    }


