
/***************************************************************************
 *   This file is part of Aspect, a simple PEC tool.                       *
 *                                                                         *
 *   Copyright (C) 2006-2007 by Wolfgang Hoffmann <woho@woho.de>           *
 *                                                                         *
 *   This program is free software, licensed under the GPL v2.             *
 *   See the file COPYING for more details.                                *
 ***************************************************************************/


#include "mcuserial.h"
#include "mcucmd.h"
#include "log.h"

#include "qextserialport.h"

#include <QtCore>




/********************************************************************/
/*!
    \class McuSerial
    \brief MCU serial communication handler

    This class encapsulates the serial port that is used to
    talk to the MCU.

    It provides methods for passing commands described by IMcuCmd
    derived classes to the MCU and filling in the received answers.
*/


McuSerial::McuSerial()
:   m_pSerPort(0),
    m_bOpen(false)
    {
    }


McuSerial::~McuSerial()
    {
    close();
    }


bool McuSerial::open(int nPort)
    {
    close();

#ifdef Q_WS_WIN
    // "COM1" ... "COM4"
    QString qsPortName = QString::fromUtf8("COM%1").arg(nPort+1);
#else
    // "/dev/ttyS0" ... "/dev/ttyS3"
    QString qsPortName = QString::fromUtf8("/dev/ttyS%1").arg(nPort);
#endif

    m_pSerPort = new QextSerialPort(qsPortName);
    m_pSerPort->setBaudRate(QextSerialPort::BAUD9600);
    m_pSerPort->setDataBits(QextSerialPort::DATA_8);
    m_pSerPort->setParity(QextSerialPort::PAR_NONE);
    m_pSerPort->setStopBits(QextSerialPort::STOP_1);
    m_pSerPort->setFlowControl(QextSerialPort::FLOW_OFF);

    m_bOpen = m_pSerPort->open();
    if (!m_bOpen)
        {
        LOG(0, "McuSerial::open(%d): failed to open port \"%s\"", nPort, LQS(qsPortName));
        return false;
        }

    return true;
    }


void McuSerial::close()
    {
    if (m_bOpen)
        m_pSerPort->close();
    m_bOpen = false;

    delete m_pSerPort;
    m_pSerPort = 0;
    }


/*!
    Send command \a pCmd over serial line to MCU and fill back
    answer into \a pCmd.

    Since the communication with QExtSerialPort seems to be very fragile
    when attempting to read more answer characters than available, an
    iterative approach is chosen: \a pCmd is queried for the expected
    answer length with IMcuCmd::receive(), then all currently available
    answer characters up to that number are read from the serial line
    and handed over to \a pCmd, again with IMcuCmd::receive(), and so on.

    \note: Care must be taken with conversions from/to QChar and QString:
    - Strings that are sent/received from the MCU must be converted
      with toLatin1()/fromLatin1()
    - Strings that come from source code must be converted with
      the editor's encoding, here fromUtf8()
    - Strings that are output to the debugging console must be
      converted with LQS() (i.e. toLocal8Bit())
*/

bool McuSerial::send(IMcuCmd *pCmd)
    {
    if (pCmd == 0)
        return false;

    if (!m_bOpen)
        {
        LOG(0, "McuSerial::send(%s): port not open", LQS(pCmd->name()));
        return false;
        }

    if (pCmd->flush())
        {
        // read any orphaned answer bytes
        if (m_pSerPort->bytesAvailable() > 0)
            {
            // orphaned answer bytes are available; flush them
            int nRemain = m_pSerPort->bytesAvailable();
            LOG(0, "McuSerial::send(%s): discarding %d orphaned answer bytes",
                LQS(pCmd->name()), nRemain);
            while (m_pSerPort->bytesAvailable() > 0)
                {
                char cAns;
                m_pSerPort->getChar(&cAns);
                LOG(5, "read: 0x%02x '%c'", (int)(unsigned char)cAns,
                    QChar::fromLatin1(cAns).isPrint()? cAns: '.');
                }
            }
        }

    QByteArray aCmd = pCmd->command();
    int nCmdLen = aCmd.length();

    if (DO_LOG(4))
        {
        QString qsSend;
        for (int nByte = 0; nByte < nCmdLen; nByte++)
            qsSend += QString("%1 ").arg(
                (uint)(unsigned char)aCmd[nByte], 2, 16, QLatin1Char('0'));
        qsSend += '"';
        for (int nByte = 0; nByte < nCmdLen; nByte++)
            {
            QChar oChar = QChar::fromLatin1(aCmd[nByte]);
            qsSend += (oChar.isPrint())? oChar: QLatin1Char('.');
            }
        qsSend += '"';
        LOG(4, "send: %s", LQS(qsSend));
        }

    for (int nByte = 0; nByte < nCmdLen; nByte++)
        {
        LOG(5, "send: 0x%02x '%c'", (int)(unsigned char)aCmd[nByte],
            QChar::fromLatin1(aCmd[nByte]).isPrint()? aCmd[nByte]: '.');
        if (!m_pSerPort->putChar(aCmd[nByte]))
            {
            LOG(0, "McuSerial::send(%s): failed to send (port \"%s\")",
                LQS(pCmd->name()), LQS(m_pSerPort->name()));
            return false;
            }
        }

    QByteArray aFullAns;
    for (int nAnsLen = pCmd->receive(QByteArray()); nAnsLen > 0; )
        {
        // query how much characters are available at the serial port.
        // if less than the expected, wait 50 ms and query again.
        // then if still less, read and return what is available
        int nBytesAvail = m_pSerPort->bytesAvailable();
        if (nAnsLen > nBytesAvail)
            {
            msleep(50);
            nBytesAvail = m_pSerPort->bytesAvailable();
            }
        if (nAnsLen > nBytesAvail)
            nAnsLen = nBytesAvail;
        QByteArray aAns;
        for (int nByte = 0; nByte < nAnsLen; nByte++)
            {
            char cAns;
            if (!m_pSerPort->getChar(&cAns))
                {
                LOG(0, "McuSerial::send(%s): failed to receive (port \"%s\")",
                    LQS(pCmd->name()), LQS(m_pSerPort->name()));
                break;
                }
            aAns += cAns;
            LOG(5, "read: 0x%02x '%c'", (int)(unsigned char)cAns,
                QChar::fromLatin1(cAns).isPrint()? cAns: '.');
            }
        aFullAns += aAns;
        nAnsLen = pCmd->receive(aAns);
        }

    if (m_pSerPort->bytesAvailable() > 0)
        {
        // futher bytes are available; this should not happen
        // if it does, output a warning and flush the data
        int nRemain = m_pSerPort->bytesAvailable();
        LOG(0, "McuSerial::send(%s): discarding %d further answer bytes",
            LQS(pCmd->name()), nRemain);
        while (m_pSerPort->bytesAvailable() > 0)
            {
            char cAns;
            m_pSerPort->getChar(&cAns);
            LOG(5, "read: 0x%02x '%c'", (int)(unsigned char)cAns,
                QChar::fromLatin1(cAns).isPrint()? cAns: '.');
            }
        }

    if (aFullAns.length() == 0)
        {
        // if no answer is expected, sleep a bit to prevent
        // flooding the MCU with commands
        msleep(20);
        }

    if (DO_LOG(4))
        {
        int nAnsRead = aFullAns.length();
        QString qsRecv;
        for (int nByte = 0; nByte < nAnsRead; nByte++)
            qsRecv += QString("%1 ").arg(
                (uint)(unsigned char)aFullAns[nByte], 2, 16, QLatin1Char('0'));
        qsRecv += '"';
        for (int nByte = 0; nByte < nAnsRead; nByte++)
            {
            QChar oChar = QChar::fromLatin1(aFullAns[nByte]);
            qsRecv += (oChar.isPrint())? oChar: QLatin1Char('.');
            }
        qsRecv += '"';
        LOG(4, "recv: %s", LQS(qsRecv));
        }

    return true;
    }


/*!
    Sleep 50 milliseconds

    \note: Use with care: will stall the GUI while sleeping when called from GUI thread
*/

void McuSerial::msleep(int nMsecs) const
    {
#if 0
    // QThread::msleep is protected ...
    QThread::msleep(50);
#else
    // ... so trick around it by timing out on a fake QWaitCondition.
    // Yes, we're stalling the GUI here, but better that then
    // firing concurrent commands at the MCU
    QMutex oLock;
    oLock.lock();
    QWaitCondition oWait;
    oWait.wait(&oLock, nMsecs);
    oLock.unlock();
#endif
    }

